/****************************************************************************
 *																			*
 *					RFM12 basierte, drahtlos RS232 Verbindung				*
 *																			*
 *								Version 0.9									*
 *																			*
 *								 by Benedikt								*
 *																			*
 *						Email:	benedikt83 t gmx.net						*
 *																			*
 ****************************************************************************
 *																			*
 *	Die Software darf frei kopiert und verndert werden, solange sie nicht	*
 *	ohne meine Erlaubnis fr kommerzielle Zwecke eingesetzt wird.			*
 *																			*
 ***************************************************************************/

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include <stdlib.h>
#include <portbits.h>
#include "global.h"
#include "uart.h"
#include "rf12.h"
#include "leds.h"

#define F_CPU 10000000UL
#include <util/delay.h>

#define RF_BAUDRATE		35000		// Baudrate des RFM12
#define UART_BAUDRATE	19200

#define MAX_BUF			128			// Paket Gre in Bytes (maximal 250)
#define TX_TIMEOUT		200			// Maximale Wartezeit auf Daten in ms (max 255)

#define ANSWER_TIMEOUT	10			// Maximale Wartezeit auf die Besttigung der Daten in ms (max 255)

#define RETRY			10			// Maximale Anzahl an Sendeversuchen

volatile unsigned char delaycnt;
unsigned char rxbuf[MAX_BUF+2];		// Puffer fr empfangene Daten
unsigned char txbuf[MAX_BUF+2];		// Puffer fr zu sendende Daten
unsigned char flags, tx_cnt, tx_id, retrans_cnt;

#define WAITFORACK		1
#define SENDACK			2

void rx_packet(void);
void tx_packet(unsigned char status);

int main(void)
{	PORTD=254;
	DDRD=254;
	DDRC=255;
    uart_init(UART_BAUD_SELECT(UART_BAUDRATE, F_CPU));
	rf12_init();					// ein paar Register setzen (z.B. CLK auf 10MHz)
	rf12_config(RF_BAUDRATE, 1, 0, QUIET);	// Baudrate, Kanal (0-3), Leistung (0=max, 7=min), Umgebungsbedingungen (QUIET, NOISY)

	TCCR1A=0;
	TCCR1B=(1<<WGM12)|1;
	OCR1A=((F_CPU+500)/1000)-1;
	TIMSK=(1<<OCIE1A);

	sei();

	rf12_rxmode();

	for (;;)
	{	if (rf12_data())							// Daten im RF12 RX Puffer ?
		{	rx_packet();
		}
		if (!(flags&WAITFORACK))					// Puffer nur dann fllen, wenn die alten Daten aus dem Puffer wirklich angekommen sind
		{	if (uart_data())						// Daten im UART RX Puffer ?
			{	if (!tx_cnt)
					delaycnt=TX_TIMEOUT;			// Nach dem ersten Byte: timeout starten
				txbuf[tx_cnt++]=uart_getchar();
			}
			if (((tx_cnt)&&(delaycnt==0))||(tx_cnt>=MAX_BUF)) // Puffer voll, oder timeout seit erstem Byte im Puffer vorbei ?  -> senden
			{	tx_packet(0);
				retrans_cnt=RETRY;						// Retry Counter neu starten
			}
		}
		else if (delaycnt==0)						// Timeout: Daten nochmal senden
		{	LED_RETRANS=1;
			flags&=~WAITFORACK;						// Retransmit erlauben
			if (retrans_cnt)
			{	retrans_cnt--;
            	tx_packet(0);
			}
			else									// Versuche abgelaufen
			{	LED_ERR=1;							// -> Fehler LED an
				tx_cnt=0;							// -> Daten verwerfen
				tx_id++;
				delaycnt=1;
			}
		}
	}
}

void rx_packet(void)
{	static unsigned char rx_lastid=255;
	unsigned char rx_cnt,i, rx_id, status;
	rx_cnt=rf12_rxdata(rxbuf, &status, &rx_id);	// komplettes Paket empfangen
	if (rx_cnt<=MAX_BUF)						// Daten gltig (d.h. kein CRC Fehler) ?
	{	if (status&RECEIVED_OK)					// Empfangsbesttigung ?
		{	flags&=~WAITFORACK;					// -> "Warten auf Besttigung"-Flag lschen
			tx_cnt=0;							// -> Daten als gesendet markieren
			tx_id++;
			LED_RETRANS=0;
			LED_ERR=0;
		}
		if (rx_cnt)								// Daten empfangen
		{	tx_packet(RECEIVED_OK);				// Empfangsbesttigung senden
			if (rx_id!=rx_lastid)				// Handelt es sich um neue Daten ?
			{	for (i=0; i<rx_cnt; i++)		// Daten weiterleiten
					uart_putc(rxbuf[i]);
				rx_lastid=rx_id;				// Aktuelle ID speichern
			}
			LED_RETRANS=0;
			LED_ERR=0;
    	}
	}
	else
	{	LED_RETRANS=1;
		tx_packet(RECEIVED_FAIL);				// "Daten beschdigt"-Kennung senden
	}
}

void tx_packet(unsigned char status)
{
	rf12_stoprx();								// auf TX umschalten			
	_delay_ms(1);
	if ((flags&WAITFORACK)||(tx_cnt==0))		// es wird noch auf eine antwort vom Paket gewartet oder es sind keine Daten zu senden
	    rf12_txdata(txbuf, 0, status, 0);		// -> kein neues Paket, sondern nur Status senden
	else
	{	rf12_txdata(txbuf, tx_cnt, status, tx_id); // komplettes Paket senden
		flags|=WAITFORACK;						// auf ACK warten
		delaycnt=ANSWER_TIMEOUT;				// Timeout Counter neu starten
	}
	rf12_rxmode();								// wieder auf RX umschalten
}

ISR(SIG_OUTPUT_COMPARE1A)						// 1kHz Interrupt
{	
	if (delaycnt)
		delaycnt--;
}



